import logging
import pywikibot
from pathlib import Path
from pypypan.excel import read_pattypan_input, CommonsItem

site = pywikibot.Site('commons:commons')
test_site = pywikibot.Site('commons:test')


def sanitize_title(title: str) -> str:
    # Remove double apostrophes: https://commons.wikimedia.org/wiki/MediaWiki:Titleblacklist-custom-double-apostrophe
    return title.replace("''", "'")


def upload_image(item: CommonsItem, site: pywikibot.Site, update_existing: bool = False, dry_run: bool = False) -> False:
    imagepage = pywikibot.FilePage(site, sanitize_title(item.title))
    imagepage.text = item.description
    try:
        if imagepage.exists():
            if update_existing:
                logging.info(f"Updating description of {item.path.name}")
                if not dry_run:
                    imagepage.save("update")
                return True
            else:
                logging.info("File {} exists on {}, updating".format(
                    item.path.name, str(site)))
                return False
        else:
            if dry_run:
                return True
            else:
                return site.upload(filepage=imagepage, source_filename=item.path, ignore_warnings=True)

    except Exception as e:
        logging.exception(e)
        raise e


def upload_pattypan_excel(filename: Path, update_existing: bool = False, max_uploads: int = -1,
                          use_test_commons: bool = True, dry_run: bool = False, allow_missing_files: bool = False):
    try:
        items = read_pattypan_input(filename, allow_missing_files=allow_missing_files)
    except Exception as e:
        logging.error(e)
        return
    if use_test_commons:
        site = pywikibot.Site('commons:test')
    else:
        site = pywikibot.Site('commons:commons')
    n = 0
    for item in items:
        logging.info(f"Uploading {item.title} to {site}")
        if upload_image(item, update_existing=update_existing, site=site, dry_run=dry_run):
            n = n+1
        else:
            logging.error(f"Failed uploading {item.title} / {item.path.name}")
        if n >= max_uploads and max_uploads > 0:
            logging.info(
                "Did {} out of {} maximum uploads".format(n, max_uploads))
            return



from pathlib import Path
import pandas as pd
from typing import List
from dataclasses import dataclass
import logging


@dataclass
class CommonsItem:
    title: str
    description: str
    path: Path


def generate_pattypan_excel(excel_file: Path, image_dir: Path):
    """
    Function that generates a Pattypan Excel file that contains the images from image_dir
    """
    pass


def read_pattypan_input(filename: Path, allow_missing_files: bool = False) -> List[CommonsItem]:
    """
    Reads an Excel file following the Pattypan format:
        (https://commons.wikimedia.org/wiki/Commons:Pattypan/Simple_manual)
    This should have two worksheets:
    - 'Template', containing a single cell A1 that holds the template used
      to generate the description for the Commons page
    - 'Data', containing rows with all items to upload.
      There are two mandatory columns:
        - name: the name (title) of the item
        - path: the path to the file to upload
      All other columns are optional and free to choose,
      but they should match a placeholder in the beforementioned template.
      E.g., a column 'creator' will be used to fill the ${creator} placeholder in the template

    The functions returns a list of CommonsItem objects,
    which hold a title, a description and a path to the file to upload.
    """
    if not filename.exists():
        raise FileNotFoundError(f"File {filename} does not exist")

    try:
        data = pd.read_excel(filename, sheet_name='Data')
    except ValueError:
        raise ValueError(f"Sheet 'Data' not found in {filename}")

    try:
        tpl = pd.read_excel(filename, sheet_name='Template')
    except ValueError:
        raise ValueError(f"Sheet 'Template' not found in {filename}")

    try:
        tpl = tpl.columns[0]
        # Replace brackets
        tpl = tpl.replace("{{", "{{{{")
        tpl = tpl.replace("}}", "}}}}")
    except ValueError:
        raise ValueError(f"Cell A1 not found in {filename}")

    allowed_missing_columns = ['name', 'path']
    for col in data.columns:
        if f"${{{col}}}" not in tpl and col not in allowed_missing_columns:
            raise ValueError(f"Column {col} not used in template")

    items = []
    for i, row in data.iterrows():
        """
        There are three options for the file path:
        - Absolute
        - Relative to Pattypan excel location
        - Relative to current working dir
        All of these are checked (in this order) and if one of them exists, it is used.
        """
        # TODO: check if this is ok
        # Convert backslashes to forward, since both POSIX and Windows handle the latter well
        rawpath = Path(row['path'].replace('\\', '/'))
        if rawpath.is_absolute():
            path = rawpath
        elif (filename.parent / rawpath).exists():
            path = filename.parent / rawpath
        elif (Path.cwd() / rawpath).exists():
            path = Path.cwd() / rawpath
        else:
            path = rawpath
        if not path.exists():
            if allow_missing_files:
                logging.error(f"File {row['path']} does not exist")
            else:
                raise FileNotFoundError(f"File {row['path']} does not exist")
        else:
            logging.debug("File exists: {}".format(path))

        tpl_values = {}
        for col in data.columns:
            if col in allowed_missing_columns:
                continue
            if col == 'category':
                # Special case: expand categories
                if ";" in row[col]:
                    categories = row[col].split(";")
                    tpl_values[col] = "]]\r\n[[Category:".join(categories)
                else:
                    tpl_values[col] = row[col]
            else:
                tpl_values[col] = row[col]
        items.append(CommonsItem(path=path,
                                 description=tpl.replace(
                                     "${", "{").format(**tpl_values),
                                 title=row['name']))
    return items



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypypan']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.4.0,<2.0.0',
 'pywikibot>=6.6.3,<7.0.0',
 'requests-oauthlib>=1.3.0,<2.0.0',
 'typer>=0.4.0,<0.5.0',
 'wikitextparser>=0.48.0,<0.49.0',
 'xlrd>=2.0.1,<3.0.0']

entry_points = \
{'console_scripts': ['pypypan = pypypan.main:main']}

setup_kwargs = {
    'name': 'pypypan',
    'version': '0.8.0',
    'description': '',
    'long_description': None,
    'author': 'Ate te Voortwis',
    'author_email': 'atv@allseas.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from pathlib import Path
from pypypan.upload import upload_pattypan_excel
from pypypan.excel import read_pattypan_input, generate_pattypan_excel
import typer
import pywikibot  # noqa: F401
import logging


# TODO: Cleanup logging, put in config
logging.getLogger().setLevel(logging.INFO)
logger = logging.getLogger("main")
formatter = logging.Formatter('%(asctime)s | %(levelname)s | %(message)s')
# create file handler which logs even debug messages
fh = logging.FileHandler('pypypan.log')
fh.setFormatter(formatter)
fh.setLevel(logging.DEBUG)
logging.getLogger().addHandler(fh)

stdout_handler = logging.StreamHandler()
stdout_handler.setFormatter(formatter)
stdout_handler.setFormatter(formatter)
logging.getLogger().addHandler(stdout_handler)

logger.setLevel(logging.DEBUG)
logging.getLogger("pywiki").setLevel(logging.ERROR)
app = typer.Typer()


@app.command()
def upload(excel_file: Path, update_existing: bool = False, dry_run: bool = False,
           test_one: bool = False, use_test_commons: bool = False, allow_missing_files: bool = False):
    """
    Upload an excel file that is filled according to the Pattypan format.
    """
    if test_one:
        upload_pattypan_excel(excel_file, update_existing=update_existing, max_uploads=1,
                              allow_missing_files=allow_missing_files, use_test_commons=use_test_commons, dry_run=dry_run)
    else:
        upload_pattypan_excel(excel_file, update_existing=update_existing,
                              allow_missing_files=allow_missing_files, use_test_commons=use_test_commons, dry_run=dry_run)


@app.command()
def generate_excel(excel_file: Path = Path("./data.xls"), image_dir: Path = Path(".")):
    """
    Generate an excel file that is filled according to the Pattypan format.
    image_dir points to a directory containing the images to use.
    """
    if excel_file.exists():
        raise(FileExistsError(f"Excel file {excel_file} already exists"))

    generate_pattypan_excel(excel_file, image_dir)


@app.command()
def test_template(excel_file: Path):
    """
    Parse a template and use the first file to generate the description.
    For testing purposes, so files listed in the Excel are not checked for existence.
    """
    items = read_pattypan_input(excel_file, allow_missing_files=True)
    print(items[0].description)


@app.command()
def version():
    """
    Print the pypypan version number.
    Run pip3 install --upgrade pypypan to udpate.
    """
    import importlib.metadata
    print(importlib.metadata.version('pypypan'))

@app.callback()
def mainapp(verbose: bool = False):
    """
    Upload images to Wikimedia Commons from a simple Excel file.
    """
    if verbose:
        logging.getLogger().setLevel(logging.DEBUG)
        logger.debug("Enabled verbose logging")

def main():
    app()

if __name__ == "__main__":
    main()
